# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import (
    SupportsIntegralOps,
    SupportsIntegralOpsSCT,
    SupportsIntegralOpsSCU,
    SupportsIntegralPow,
    SupportsIntegralPowSCT,
    SupportsIntegralPowSCU,
)

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def ops_func(arg: SupportsIntegralOps):
    assert isinstance(arg, SupportsIntegralOps), f"{arg!r}"


@beartype
def ops_func_t(arg: SupportsIntegralOpsSCU):
    assert isinstance(arg, SupportsIntegralOpsSCT), f"{arg!r}"


@beartype
def pow_func(arg: SupportsIntegralPow):
    assert isinstance(arg, SupportsIntegralPow), f"{arg!r}"


@beartype
def pow_func_t(arg: SupportsIntegralPowSCU):
    assert isinstance(arg, SupportsIntegralPowSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_supports_integral_ops_pow() -> None:
    bool_val: SupportsIntegralOps = True
    int_val: SupportsIntegralOps = -273
    nw_val: SupportsIntegralOps = Numberwang(-273)
    nwd_val: SupportsIntegralOps = NumberwangDerived(-273)
    nwr_val: SupportsIntegralOps = NumberwangRegistered(-273)
    _: SupportsIntegralPow
    _ = True
    _ = -273
    _ = Numberwang(-273)
    _ = NumberwangDerived(-273)
    _ = NumberwangRegistered(-273)

    for good_val in (
        bool_val,
        int_val,
        nw_val,
        nwd_val,
        nwr_val,
    ):
        assert isinstance(good_val, SupportsIntegralOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralOpsSCT), f"{good_val!r}"
        assert good_val >> 0 == good_val, f"{good_val!r}"
        assert good_val << 0 == good_val, f"{good_val!r}"
        assert good_val & 0 == 0, f"{good_val!r}"
        assert good_val | 0 == good_val, f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"

    for bad_val in (
        -273.15,
        complex(-273.15),
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273.15",
    ):
        assert not isinstance(bad_val, SupportsIntegralOps), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsIntegralOpsSCT), f"{bad_val!r}"


def test_supports_integral_ops_pow_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
    ):
        ops_func(cast(SupportsIntegralOps, good_val))
        ops_func_t(cast(SupportsIntegralOpsSCU, good_val))

    for bad_val in (
        -273.15,
        complex(-273.15),
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273.15",
    ):
        with pytest.raises(roar.BeartypeException):
            ops_func(cast(SupportsIntegralOps, bad_val))

        with pytest.raises(roar.BeartypeException):
            ops_func_t(cast(SupportsIntegralOpsSCU, bad_val))


def test_supports_integral_ops_pow_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsIntegralOps = numpy.uint8(2)
    uint16_val: SupportsIntegralOps = numpy.uint16(273)
    uint32_val: SupportsIntegralOps = numpy.uint32(273)
    uint64_val: SupportsIntegralOps = numpy.uint64(273)
    int8_val: SupportsIntegralOps = numpy.int8(-2)
    int16_val: SupportsIntegralOps = numpy.int16(-273)
    int32_val: SupportsIntegralOps = numpy.int32(-273)
    int64_val: SupportsIntegralOps = numpy.int64(-273)
    float16_val: SupportsIntegralOps = numpy.float16(-1.8)
    float32_val: SupportsIntegralOps = numpy.float32(-273.15)
    float64_val: SupportsIntegralOps = numpy.float64(-273.15)
    float128_val: SupportsIntegralOps = numpy.float128(-273.15)
    _: SupportsIntegralPow
    _ = numpy.uint8(2)
    _ = numpy.uint16(273)
    _ = numpy.uint32(273)
    _ = numpy.uint64(273)
    _ = numpy.int8(-2)
    _ = numpy.int16(-273)
    _ = numpy.int32(-273)
    _ = numpy.int64(-273)
    _ = numpy.float16(-1.8)
    _ = numpy.float32(-273.15)
    _ = numpy.float64(-273.15)
    _ = numpy.float128(-273.15)
    _ = numpy.csingle(-273.15)
    _ = numpy.cdouble(-273.15)
    _ = numpy.clongdouble(-273.15)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
    ):
        assert isinstance(good_val, SupportsIntegralOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralOpsSCT), f"{good_val!r}"

        # I have no idea why numpy.uint64 is special in this regard
        if isinstance(good_val, numpy.uint64):
            assert good_val >> type(good_val)(0) == good_val, f"{good_val!r}"
            assert good_val << type(good_val)(0) == good_val, f"{good_val!r}"
            assert good_val & type(good_val)(0) == type(good_val)(0), f"{good_val!r}"
            assert good_val | type(good_val)(0) == good_val, f"{good_val!r}"
        else:
            assert good_val >> 0 == good_val, f"{good_val!r}"
            assert good_val << 0 == good_val, f"{good_val!r}"
            assert good_val & 0 == 0, f"{good_val!r}"
            assert good_val | 0 == good_val, f"{good_val!r}"

        assert isinstance(good_val, SupportsIntegralPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"

    for lying_val in (
        float16_val,
        float32_val,
        float64_val,
        float128_val,
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert isinstance(good_val, SupportsIntegralOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralOpsSCT), f"{good_val!r}"

        # These have, but don't implement these functions
        # TODO(posita): Can we fix this?
        with pytest.raises(TypeError):
            lying_val << 0

        with pytest.raises(TypeError):
            lying_val >> 0

        with pytest.raises(TypeError):
            lying_val & 0

        with pytest.raises(TypeError):
            lying_val | 0

        assert isinstance(lying_val, SupportsIntegralPow), f"{lying_val!r}"
        assert isinstance(lying_val, SupportsIntegralPowSCT), f"{lying_val!r}"
        assert lying_val ** 1 == lying_val, f"{lying_val!r}"


def test_supports_integral_ops_pow_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
    ):
        ops_func(cast(SupportsIntegralOps, good_val))
        ops_func_t(cast(SupportsIntegralOpsSCU, good_val))

    for lying_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        ops_func(cast(SupportsIntegralOps, good_val))
        ops_func_t(cast(SupportsIntegralOpsSCU, good_val))


def test_supports_integral_ops_pow_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    integral_val: SupportsIntegralOps = sympy.Integer(-273)
    sym_val: SupportsIntegralOps = sympy.symbols("x")
    _: SupportsIntegralPow
    _ = sympy.Integer(-273)
    _ = sympy.symbols("x")

    for good_val in (integral_val,):
        assert isinstance(good_val, SupportsIntegralOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralOpsSCT), f"{good_val!r}"
        assert good_val >> 0 == good_val, f"{good_val!r}"
        assert good_val << 0 == good_val, f"{good_val!r}"
        assert good_val & 0 == 0, f"{good_val!r}"
        assert good_val | 0 == good_val, f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"

    for lying_val in (sym_val,):
        assert isinstance(good_val, SupportsIntegralOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsIntegralOpsSCT), f"{good_val!r}"

        # Relationals have, but don't implement this function
        # TODO(posita): Can we fix this?
        with pytest.raises(TypeError):
            lying_val << 0

    for bad_val in (
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
    ):
        assert not isinstance(bad_val, SupportsIntegralOps), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsIntegralOpsSCT), f"{bad_val!r}"


def test_supports_integral_ops_pow_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (sympy.Integer(-273),):
        ops_func(cast(SupportsIntegralOps, good_val))
        ops_func_t(cast(SupportsIntegralOpsSCU, good_val))

    for lying_val in (sympy.symbols("x"),):
        ops_func(cast(SupportsIntegralOps, lying_val))
        ops_func_t(cast(SupportsIntegralOpsSCU, lying_val))

    for bad_val in (
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
    ):
        with pytest.raises(roar.BeartypeException):
            ops_func(cast(SupportsIntegralOps, bad_val))

        with pytest.raises(roar.BeartypeException):
            ops_func_t(cast(SupportsIntegralOpsSCU, bad_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import os
import sys
import warnings
from typing import TypeVar, cast

__all__ = ("beartype",)


# ---- Types ---------------------------------------------------------------------------


_T = TypeVar("_T")

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


# ---- Functions -----------------------------------------------------------------------


def identity(__: _T) -> _T:
    return __


# ---- Initialization ------------------------------------------------------------------


class _DecoratorT(Protocol):
    def __call__(self, __: _T) -> _T:
        ...


beartype: _DecoratorT = identity

_NUMERARY_BEARTYPE = os.environ.get("NUMERARY_BEARTYPE", "on")
_truthy = ("on", "t", "true", "yes")
_falsy = ("off", "f", "false", "no")
_use_beartype_if_available: bool

try:
    _use_beartype_if_available = bool(int(_NUMERARY_BEARTYPE))
except ValueError:
    if _NUMERARY_BEARTYPE.lower() in _truthy:
        _use_beartype_if_available = True
    elif _NUMERARY_BEARTYPE.lower() in _falsy:
        _use_beartype_if_available = False
    else:
        raise EnvironmentError(
            f"""unrecognized value ({_NUMERARY_BEARTYPE}) for NUMERARY_BEARTYPE environment variable (should be "{'", "'.join(_truthy + _falsy)}", or an integer)"""
        )

if _use_beartype_if_available:
    try:
        import beartype as _beartype

        if _beartype.__version_info__ >= (0, 8):
            beartype = cast(_DecoratorT, _beartype.beartype)
        else:
            warnings.warn(
                f"beartype>=0.8 required, but beartype=={_beartype.__version__} found; disabled"
            )
    except ImportError:
        pass



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from math import ceil, floor, trunc
from numbers import Integral, Real
from operator import (
    __abs__,
    __add__,
    __and__,
    __eq__,
    __floordiv__,
    __ge__,
    __gt__,
    __invert__,
    __le__,
    __lshift__,
    __lt__,
    __mod__,
    __mul__,
    __ne__,
    __neg__,
    __or__,
    __pos__,
    __pow__,
    __rshift__,
    __sub__,
    __truediv__,
    __xor__,
)
from typing import Iterable, Optional, Union, overload

from numerary.bt import beartype
from numerary.types import SupportsIntSCU

__all__ = (
    "Numberwang",
    "NumberwangDerived",
    "NumberwangRegistered",
    "Wangernumb",
    "WangernumbDerived",
    "WangernumbRegistered",
)


# ---- Types ---------------------------------------------------------------------------


_IntegralT = Union[int, "NumberwangBase", Integral]
_RealT = Union[float, "Wangernumb", Real]


# ---- Classes -------------------------------------------------------------------------


class NumberwangBase:
    __slots__: Union[str, Iterable[str]] = ("val",)

    @beartype
    def __init__(self, arg: SupportsIntSCU = 0):
        self.val: int = int(arg)

    @beartype
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.val})"

    @beartype
    def __lt__(self, other) -> bool:
        return __lt__(self.val, other)

    @beartype
    def __le__(self, other) -> bool:
        return __le__(self.val, other)

    @beartype
    def __eq__(self, other) -> bool:
        return __eq__(self.val, other)

    @beartype
    def __ne__(self, other) -> bool:
        return __ne__(self.val, other)

    @beartype
    def __ge__(self, other) -> bool:
        return __ge__(self.val, other)

    @beartype
    def __gt__(self, other) -> bool:
        return __gt__(self.val, other)

    @beartype
    def __hash__(self) -> int:
        return hash((type(self).__name__, self.val))

    @overload
    def __add__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __add__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __add__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __add__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__add__(self.val, other))
        elif isinstance(other, float):
            val = __add__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __add__(self.val, other)

    @overload
    def __radd__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __radd__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __radd__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __radd__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__add__(other, self.val))
        elif isinstance(other, float):
            val = __add__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __add__(other, self.val)

    @overload
    def __sub__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __sub__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __sub__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __sub__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__sub__(self.val, other))
        elif isinstance(other, float):
            val = __sub__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __sub__(self.val, other)

    @overload
    def __rsub__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __rsub__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rsub__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rsub__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__sub__(other, self.val))
        elif isinstance(other, float):
            val = __sub__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __sub__(other, self.val)

    @overload
    def __mul__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __mul__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __mul__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __mul__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__mul__(self.val, other))
        elif isinstance(other, float):
            val = __mul__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __mul__(self.val, other)

    @overload
    def __rmul__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __rmul__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rmul__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rmul__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__mul__(other, self.val))
        elif isinstance(other, float):
            val = __mul__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __mul__(other, self.val)

    @overload
    def __truediv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __truediv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __truediv__(self, other):
        if isinstance(other, (float, NumberwangBase, Wangernumb, Real)):
            val = __truediv__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __truediv__(self.val, other)

    @overload
    def __rtruediv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rtruediv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rtruediv__(self, other):
        if isinstance(other, (float, NumberwangBase, Wangernumb, Real)):
            val = __truediv__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __truediv__(other, self.val)

    @overload
    def __floordiv__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __floordiv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __floordiv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __floordiv__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__floordiv__(self.val, other))
        elif isinstance(other, float):
            val = __floordiv__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __floordiv__(self.val, other)

    @overload
    def __rfloordiv__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __rfloordiv__(self, other: _RealT) -> _RealT:  # type: ignore
        ...

    @overload
    def __rfloordiv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rfloordiv__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__floordiv__(other, self.val))
        elif isinstance(other, float):
            val = __floordiv__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __floordiv__(other, self.val)

    @overload
    def __mod__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __mod__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __mod__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __mod__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__mod__(self.val, other))
        elif isinstance(other, float):
            val = __mod__(self.val, other)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __mod__(self.val, other)

    @overload
    def __rmod__(self, other: _IntegralT) -> _IntegralT:  # type: ignore
        ...

    @overload
    def __rmod__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rmod__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rmod__(self, other):
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__mod__(other, self.val))
        elif isinstance(other, float):
            val = __mod__(other, self.val)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return __mod__(other, self.val)

    @overload
    def __pow__(  # type: ignore
        self, other: _IntegralT, modulo: Optional[_IntegralT] = None
    ) -> _IntegralT:
        ...

    @overload
    def __pow__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __pow__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __pow__(self, other, modulo=None):
        if isinstance(other, (NumberwangBase, Integral)):
            return type(self)(pow(self.val, int(other), modulo))
        elif isinstance(other, float):
            val = pow(self.val, float(other), modulo)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return pow(self.val, other, modulo)

    @overload
    def __rpow__(  # type: ignore
        self, other: _IntegralT, modulo: Optional[_IntegralT] = None
    ) -> _IntegralT:
        ...

    @overload
    def __rpow__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rpow__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rpow__(self, other, modulo=None):
        if isinstance(other, (NumberwangBase, Integral)):
            return type(self)(pow(int(other), self.val, modulo))
        elif isinstance(other, float):
            val = pow(float(other), self.val, modulo)

            if isinstance(self, NumberwangDerived):
                return WangernumbDerived(val)
            elif isinstance(self, NumberwangRegistered):
                return WangernumbRegistered(val)
            else:
                return Wangernumb(val)
        else:
            return pow(other, self.val, modulo)

    @beartype
    def __lshift__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__lshift__(self.val, int(other)))
        elif isinstance(other, Integral):
            return __lshift__(self.val, other)
        else:
            return NotImplemented

    @beartype
    def __rlshift__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__lshift__(int(other), self.val))
        elif isinstance(other, Integral):
            return __lshift__(other, self.val)
        else:
            return NotImplemented

    @beartype
    def __rshift__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__rshift__(self.val, other))
        elif isinstance(other, Integral):
            return __rshift__(self.val, other)
        else:
            return NotImplemented

    @beartype
    def __rrshift__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__rshift__(other, self.val))
        elif isinstance(other, Integral):
            return __rshift__(other, self.val)
        else:
            return NotImplemented

    @beartype
    def __and__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__and__(self.val, other))
        elif isinstance(other, Integral):
            return __and__(self.val, other)
        else:
            return NotImplemented

    @beartype
    def __rand__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__and__(other, self.val))
        elif isinstance(other, Integral):
            return __and__(other, self.val)
        else:
            return NotImplemented

    @beartype
    def __xor__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__xor__(self.val, other))
        elif isinstance(other, Integral):
            return __xor__(self.val, other)
        else:
            return NotImplemented

    @beartype
    def __rxor__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__xor__(other, self.val))
        elif isinstance(other, Integral):
            return __xor__(other, self.val)
        else:
            return NotImplemented

    @beartype
    def __or__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__or__(self.val, other))
        elif isinstance(other, Integral):
            return __or__(self.val, other)
        else:
            return NotImplemented

    @beartype
    def __ror__(self, other: _IntegralT) -> _IntegralT:
        if isinstance(other, (int, NumberwangBase)):
            return type(self)(__or__(other, self.val))
        elif isinstance(other, Integral):
            return __or__(other, self.val)
        else:
            return NotImplemented

    @beartype
    def __neg__(self) -> NumberwangBase:
        return type(self)(__neg__(self.val))

    @beartype
    def __pos__(self) -> NumberwangBase:
        return type(self)(__pos__(self.val))

    @beartype
    def __abs__(self) -> NumberwangBase:
        return type(self)(__abs__(self.val))

    @beartype
    def __invert__(self) -> NumberwangBase:
        return type(self)(__invert__(self.val))

    @beartype
    def __int__(self) -> int:
        return self.val

    @beartype
    def __round__(self, ndigits: Optional[_IntegralT] = None) -> int:
        if ndigits is None:
            return round(self.val)
        else:
            return round(self.val, int(ndigits))

    @beartype
    def __trunc__(self) -> int:
        return trunc(self.val)

    @beartype
    def __floor__(self) -> int:
        return floor(self.val)

    @beartype
    def __ceil__(self) -> int:
        return ceil(self.val)

    @property
    def numerator(self) -> int:
        return self.val

    @property
    def denominator(self) -> int:
        return 1


assert not issubclass(NumberwangBase, Real)
assert not issubclass(NumberwangBase, Integral)


class Numberwang(NumberwangBase):
    __slots__: Union[str, Iterable[str]] = ()

    @beartype
    def __float__(self) -> float:
        return float(int(self))

    @beartype
    def __index__(self) -> int:
        return int(self)


assert not issubclass(Numberwang, Real)
assert not issubclass(Numberwang, Integral)


class NumberwangRegistered(Numberwang):
    __slots__: Union[str, Iterable[str]] = ()


assert not issubclass(NumberwangRegistered, Real)
assert not issubclass(NumberwangRegistered, Integral)

Integral.register(NumberwangRegistered)
assert issubclass(NumberwangRegistered, Real)
assert issubclass(NumberwangRegistered, Integral)


class NumberwangDerived(NumberwangBase, Integral):  # type: ignore
    __slots__: Union[str, Iterable[str]] = ()


assert issubclass(NumberwangDerived, Real)
assert issubclass(NumberwangDerived, Integral)

assert not issubclass(NumberwangBase, Real)
assert not issubclass(NumberwangBase, Integral)
assert not issubclass(Numberwang, Real)
assert not issubclass(Numberwang, Integral)


class Wangernumb:
    __slots__: Union[str, Iterable[str]] = ("val",)

    @beartype
    def __init__(self, arg: _RealT = 0):
        self.val = float(arg)

    @beartype
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.val})"

    @beartype
    def __lt__(self, other: _RealT) -> bool:
        return __lt__(self.val, other)

    @beartype
    def __le__(self, other: _RealT) -> bool:
        return __le__(self.val, other)

    @beartype
    def __eq__(self, other) -> bool:
        return __eq__(self.val, other)

    @beartype
    def __ne__(self, other) -> bool:
        return __ne__(self.val, other)

    @beartype
    def __ge__(self, other: _RealT) -> bool:
        return __ge__(self.val, other)

    @beartype
    def __gt__(self, other: _RealT) -> bool:
        return __gt__(self.val, other)

    @beartype
    def __hash__(self) -> int:
        return hash(self.val)

    @overload
    def __add__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __add__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __add__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__add__(self.val, other))
        else:
            return __add__(self.val, other)

    @overload
    def __radd__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __radd__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __radd__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__add__(other, self.val))
        else:
            return __add__(other, self.val)

    @overload
    def __sub__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __sub__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __sub__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__sub__(self.val, other))
        else:
            return __sub__(self.val, other)

    @overload
    def __rsub__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rsub__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rsub__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__sub__(other, self.val))
        else:
            return __sub__(other, self.val)

    @overload
    def __mul__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __mul__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __mul__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__mul__(self.val, other))
        else:
            return __mul__(self.val, other)

    @overload
    def __rmul__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rmul__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rmul__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__mul__(other, self.val))
        else:
            return __mul__(other, self.val)

    @overload
    def __truediv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __truediv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __truediv__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__truediv__(self.val, other))
        else:
            return __truediv__(self.val, other)

    @overload
    def __rtruediv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rtruediv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rtruediv__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__truediv__(other, self.val))
        else:
            return __truediv__(other, self.val)

    @overload
    def __floordiv__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __floordiv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __floordiv__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__floordiv__(self.val, other))
        else:
            return __floordiv__(self.val, other)

    @overload
    def __rfloordiv__(self, other: _RealT) -> _RealT:  # type: ignore
        ...

    @overload
    def __rfloordiv__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rfloordiv__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__floordiv__(other, self.val))
        else:
            return __floordiv__(other, self.val)

    @overload
    def __mod__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __mod__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __mod__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__mod__(self.val, other))
        else:
            return __mod__(self.val, other)

    @overload
    def __rmod__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rmod__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rmod__(self, other):
        if isinstance(other, (float, Wangernumb)):
            return type(self)(__mod__(other, self.val))
        else:
            return __mod__(other, self.val)

    @overload
    def __pow__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __pow__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __pow__(self, other):
        val = __pow__(self.val, other)

        if isinstance(val, Real):
            return type(self)(val)
        else:
            return val

    @overload
    def __rpow__(self, other: _RealT) -> _RealT:
        ...

    @overload
    def __rpow__(self, other: Decimal) -> Decimal:
        ...

    @beartype
    def __rpow__(self, other):
        val = __pow__(other, self.val)

        if isinstance(val, Real):
            return type(self)(val)
        else:
            return val

    @beartype
    def __neg__(self) -> Wangernumb:
        return type(self)(__neg__(self.val))

    @beartype
    def __pos__(self) -> Wangernumb:
        return type(self)(__pos__(self.val))

    @beartype
    def __abs__(self) -> Wangernumb:
        return type(self)(__abs__(self.val))

    @beartype
    def __float__(self) -> float:
        return self.val

    @overload
    def __round__(self) -> int:
        ...

    @overload
    def __round__(self, ndigits: _IntegralT) -> float:
        ...

    @beartype
    def __round__(self, ndigits: Optional[_IntegralT] = None) -> Union[int, float]:
        if ndigits is None:
            return round(self.val)
        else:
            return round(self.val, int(ndigits))

    @beartype
    def __trunc__(self) -> int:
        return trunc(self.val)

    @beartype
    def __floor__(self) -> int:
        return floor(self.val)

    @beartype
    def __ceil__(self) -> int:
        return ceil(self.val)


assert not issubclass(Wangernumb, Real)


class WangernumbRegistered(Wangernumb):
    __slots__: Union[str, Iterable[str]] = ()


assert not issubclass(WangernumbRegistered, Real)

Real.register(WangernumbRegistered)
assert issubclass(WangernumbRegistered, Real)


class WangernumbDerived(Wangernumb, Real):  # type: ignore
    __slots__: Union[str, Iterable[str]] = ()


assert issubclass(WangernumbDerived, Real)

assert not issubclass(Wangernumb, Real)



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import (
    SupportsNumeratorDenominatorMethods,
    SupportsNumeratorDenominatorMixedSCT,
    SupportsNumeratorDenominatorMixedSCU,
    SupportsNumeratorDenominatorMixedT,
    SupportsNumeratorDenominatorMixedU,
    SupportsNumeratorDenominatorProperties,
    denominator,
    numerator,
)

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Classes -------------------------------------------------------------------------


class SageLikeRational:
    r"""
    Hack to emulate a rational similar to Sage’s implementation, where the numerator and
    denominator are exposed via methods, not properties.

    Sage’s rational implementation predates PEP 3141. Unfortunately, fixing this in Sage
    would be a breaking change. (See [this
    issue](https://trac.sagemath.org/ticket/28234) for more details.
    """

    def __init__(self, numerator: int, denominator: int = 1):
        self._numerator = numerator
        self._denominator = denominator

    def numerator(self) -> int:
        return self._numerator

    def denominator(self) -> int:
        return self._denominator


# ---- Functions -----------------------------------------------------------------------


@beartype
def func(arg: SupportsNumeratorDenominatorMixedU):
    assert isinstance(arg, SupportsNumeratorDenominatorMixedT), f"{arg!r}"


@beartype
def func_t(arg: SupportsNumeratorDenominatorMixedSCU):
    assert isinstance(arg, SupportsNumeratorDenominatorMixedSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_numerator_denominator() -> None:
    bool_val: SupportsNumeratorDenominatorProperties = True
    int_val: SupportsNumeratorDenominatorProperties = -273
    frac_val: SupportsNumeratorDenominatorProperties = Fraction(-27315, 100)
    nw_val: SupportsNumeratorDenominatorProperties = Numberwang(-273)
    nwd_val: SupportsNumeratorDenominatorProperties = NumberwangDerived(-273)
    nwr_val: SupportsNumeratorDenominatorProperties = NumberwangRegistered(-273)
    sage_val: SupportsNumeratorDenominatorMethods = SageLikeRational(-27315, 100)

    for good_val in (
        bool_val,
        int_val,
        frac_val,
        nw_val,
        nwd_val,
        nwr_val,
        sage_val,
    ):
        assert isinstance(good_val, SupportsNumeratorDenominatorMixedT), f"{good_val!r}"
        assert isinstance(
            good_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{good_val!r}"
        assert numerator(good_val), f"{good_val!r}"
        assert denominator(good_val), f"{good_val!r}"

    for bad_val in (
        -273.15,
        complex(-273.15),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedT
        ), f"{bad_val!r}"
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{bad_val!r}"


def test_numerator_denominator_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        Fraction(-27315, 100),
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
        SageLikeRational(-27315, 100),
    ):
        func(cast(SupportsNumeratorDenominatorMixedU, good_val))
        func_t(cast(SupportsNumeratorDenominatorMixedSCU, good_val))

    for bad_val in (
        -273.15,
        complex(-273.15),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsNumeratorDenominatorMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsNumeratorDenominatorMixedSCU, bad_val))


def test_numerator_denominator_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsNumeratorDenominatorProperties = numpy.uint8(2)
    uint16_val: SupportsNumeratorDenominatorProperties = numpy.uint16(273)
    uint32_val: SupportsNumeratorDenominatorProperties = numpy.uint32(273)
    uint64_val: SupportsNumeratorDenominatorProperties = numpy.uint64(273)
    int8_val: SupportsNumeratorDenominatorProperties = numpy.int8(-2)
    int16_val: SupportsNumeratorDenominatorProperties = numpy.int16(-273)
    int32_val: SupportsNumeratorDenominatorProperties = numpy.int32(-273)
    int64_val: SupportsNumeratorDenominatorProperties = numpy.int64(-273)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
    ):
        assert isinstance(good_val, SupportsNumeratorDenominatorMixedT), f"{good_val!r}"
        assert isinstance(
            good_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{good_val!r}"
        assert numerator(good_val), f"{good_val!r}"
        assert denominator(good_val), f"{good_val!r}"

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedT
        ), f"{bad_val!r}"
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{bad_val!r}"


def test_numerator_denominator_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
    ):
        func(cast(SupportsNumeratorDenominatorMixedU, good_val))
        func_t(cast(SupportsNumeratorDenominatorMixedSCU, good_val))

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsNumeratorDenominatorMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsNumeratorDenominatorMixedSCU, bad_val))


def test_numerator_denominator_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    integral_val: SupportsNumeratorDenominatorProperties = sympy.Integer(-273)
    rational_val: SupportsNumeratorDenominatorProperties = sympy.Rational(-27315, 100)

    for good_val in (
        integral_val,
        rational_val,
    ):
        assert isinstance(good_val, SupportsNumeratorDenominatorMixedT), f"{good_val!r}"
        assert isinstance(
            good_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{good_val!r}"
        assert numerator(good_val), f"{good_val!r}"
        assert denominator(good_val), f"{good_val!r}"

    for bad_val in (
        sympy.Float(-273.15),
        sympy.symbols("x"),
    ):
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedT
        ), f"{bad_val!r}"
        assert not isinstance(
            bad_val, SupportsNumeratorDenominatorMixedSCT
        ), f"{bad_val!r}"


def test_numerator_denominator_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Rational(-27315, 100),
        sympy.Integer(-273),
    ):
        func(cast(SupportsNumeratorDenominatorMixedU, good_val))
        func_t(cast(SupportsNumeratorDenominatorMixedSCU, good_val))

    for bad_val in (
        sympy.Float(-273.15),
        sympy.symbols("x"),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsNumeratorDenominatorMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsNumeratorDenominatorMixedSCU, bad_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import math
import sys
from abc import abstractmethod
from decimal import Decimal
from fractions import Fraction
from numbers import Complex, Integral, Rational, Real
from typing import Any, Dict, Iterable, Optional, Tuple, Type, TypeVar, Union

from .bt import beartype

__all__ = (
    "IntegralLike",
    "IntegralLikeSCU",
    "IntegralLikeSCT",
    "RealLike",
    "RealLikeSCU",
    "RealLikeSCT",
)


# ---- Types ---------------------------------------------------------------------------


_T_co = TypeVar("_T_co", covariant=True)
_TT = TypeVar("_TT", bound="type")

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated  # noqa: F401

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import SupportsAbs as _SupportsAbs
    from typing import SupportsComplex as _SupportsComplex
    from typing import SupportsFloat as _SupportsFloat
    from typing import SupportsIndex as _SupportsIndex
    from typing import SupportsInt as _SupportsInt
    from typing import SupportsRound as _SupportsRound
    from typing import runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

    @runtime_checkable
    class _SupportsAbs(Protocol[_T_co]):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __abs__(self) -> _T_co:
            pass

    @runtime_checkable
    class _SupportsComplex(Protocol):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __complex__(self) -> complex:
            pass

    @runtime_checkable
    class _SupportsFloat(Protocol):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __float__(self) -> float:
            pass

    @runtime_checkable
    class _SupportsIndex(Protocol):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __index__(self) -> int:
            pass

    @runtime_checkable
    class _SupportsInt(Protocol):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __int__(self) -> int:
            pass

    @runtime_checkable
    class _SupportsRound(Protocol[_T_co]):
        __slots__: Union[str, Iterable[str]] = ()

        @abstractmethod
        def __round__(self, ndigits: int = 0) -> _T_co:
            pass


_ProtocolMeta: Any = type(Protocol)


class CachingProtocolMeta(_ProtocolMeta):
    """
    Stand-in for ``#!python Protocol``’s base class that caches results of ``#!python
    __instancecheck__``, (which is otherwise [really 🤬ing
    expensive](https://github.com/python/mypy/issues/3186#issuecomment-885718629)).
    (When this was introduced, it resulted in about a 5× performance increase for
    [``dyce``](https://github.com/posita/dyce)’s unit tests, which was the only
    benchmark I had at the time.) The downside is that this will yield unpredictable
    results for objects whose methods don’t stem from any type (e.g., are assembled at
    runtime). I don’t know of any real-world case where that would be true. We’ll jump
    off that bridge when we come to it.
    """

    def __new__(
        mcls: Type[_TT],
        name: str,
        bases: Tuple[Type, ...],
        namespace: Dict[str, Any],
        **kw: Any,
    ) -> _TT:
        # See <https://github.com/python/mypy/issues/9282>
        cls = super().__new__(mcls, name, bases, namespace, **kw)  # type: ignore
        # Prefixing this class member with "_abc_" is necessary to prevent it from being
        # considered part of the Protocol. (See
        # <https://github.com/python/cpython/blob/main/Lib/typing.py>.)
        cache: Dict[Tuple[type, type], bool] = {}
        cls._abc_inst_check_cache = cache

        return cls

    def __instancecheck__(self, inst: Any) -> bool:
        inst_t = type(inst)

        if (self, inst_t) not in self._abc_inst_check_cache:
            self._abc_inst_check_cache[self, inst_t] = super().__instancecheck__(inst)

        return self._abc_inst_check_cache[self, inst_t]


def _assert_isinstance(*num_ts: type, target_t: type) -> None:
    for num_t in num_ts:
        assert isinstance(num_t(0), target_t), f"{num_t!r}, {target_t!r}"


@runtime_checkable
class SupportsAbs(
    _SupportsAbs[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsAbs.__doc__ = _SupportsAbs.__doc__
_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsAbs)
SupportsAbsSCU = Union[int, float, bool, Complex, SupportsAbs]
SupportsAbsSCT = (int, float, bool, Complex, SupportsAbs)
assert SupportsAbsSCU.__args__ == SupportsAbsSCT  # type: ignore


@runtime_checkable
class SupportsComplex(
    _SupportsComplex,
    Protocol,
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsComplex.__doc__ = _SupportsComplex.__doc__
_assert_isinstance(Decimal, Fraction, target_t=SupportsComplex)
SupportsComplexSCU = Union[Complex, SupportsComplex]
SupportsComplexSCT = (Complex, SupportsComplex)
assert SupportsComplexSCU.__args__ == SupportsComplexSCT  # type: ignore


@runtime_checkable
class SupportsFloat(
    _SupportsFloat,
    Protocol,
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsFloat.__doc__ = _SupportsFloat.__doc__
_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsFloat)
SupportsFloatSCU = Union[int, float, bool, Real, SupportsFloat]
SupportsFloatSCT = (int, float, bool, Real, SupportsFloat)
assert SupportsFloatSCU.__args__ == SupportsFloatSCT  # type: ignore


@runtime_checkable
class SupportsInt(
    _SupportsInt,
    Protocol,
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsInt.__doc__ = _SupportsInt.__doc__
_assert_isinstance(int, float, bool, target_t=SupportsInt)
SupportsIntSCU = Union[int, float, bool, Integral, SupportsInt]
SupportsIntSCT = (int, float, bool, Integral, SupportsInt)
assert SupportsIntSCU.__args__ == SupportsIntSCT  # type: ignore


@runtime_checkable
class SupportsIndex(
    _SupportsIndex,
    Protocol,
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsIndex.__doc__ = _SupportsIndex.__doc__
_assert_isinstance(int, bool, target_t=SupportsIndex)
SupportsIndexSCU = Union[int, bool, Integral, SupportsIndex]
SupportsIndexSCT = (int, bool, Integral, SupportsIndex)
assert SupportsIndexSCU.__args__ == SupportsIndexSCT  # type: ignore


@runtime_checkable
class SupportsRound(
    _SupportsRound[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    __slots__: Union[str, Iterable[str]] = ()


SupportsRound.__doc__ = _SupportsRound.__doc__
_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsRound)
SupportsRoundSCU = Union[int, float, bool, Real, SupportsRound]
SupportsRoundSCT = (int, float, bool, Real, SupportsRound)
assert SupportsRoundSCU.__args__ == SupportsRoundSCT  # type: ignore


@runtime_checkable
class SupportsConjugate(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def conjugate(self) -> Any:
        pass


_assert_isinstance(
    int, float, bool, complex, Decimal, Fraction, target_t=SupportsConjugate
)
SupportsConjugateSCU = Union[int, float, bool, complex, Complex, SupportsConjugate]
SupportsConjugateSCT = (int, float, bool, complex, Complex, SupportsConjugate)
assert SupportsConjugateSCU.__args__ == SupportsConjugateSCT  # type: ignore


@runtime_checkable
class SupportsRealImag(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @property
    def real(self) -> Any:
        pass

    @property
    def imag(self) -> Any:
        pass


_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsRealImag)
SupportsRealImagSCU = Union[int, float, bool, Complex, SupportsRealImag]
SupportsRealImagSCT = (int, float, bool, Complex, SupportsRealImag)
assert SupportsRealImagSCU.__args__ == SupportsRealImagSCT  # type: ignore


@runtime_checkable
class SupportsTrunc(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __trunc__(self) -> int:
        pass


_assert_isinstance(int, bool, float, Decimal, Fraction, target_t=SupportsTrunc)
SupportsTruncSCU = Union[int, float, bool, Real, SupportsTrunc]
SupportsTruncSCT = (int, float, bool, Real, SupportsTrunc)
assert SupportsTruncSCU.__args__ == SupportsTruncSCT  # type: ignore


@runtime_checkable
class SupportsFloor(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __floor__(self) -> int:
        pass


_assert_isinstance(int, bool, Decimal, Fraction, target_t=SupportsFloor)

if sys.version_info >= (3, 9):
    _assert_isinstance(float, target_t=SupportsFloor)

SupportsFloorSCU = Union[int, float, bool, Real, SupportsFloor]
SupportsFloorSCT = (int, float, bool, Real, SupportsFloor)
assert SupportsFloorSCU.__args__ == SupportsFloorSCT  # type: ignore


@runtime_checkable
class SupportsCeil(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __ceil__(self) -> int:
        pass


_assert_isinstance(int, bool, Decimal, Fraction, target_t=SupportsCeil)

if sys.version_info >= (3, 9):
    _assert_isinstance(float, target_t=SupportsCeil)

SupportsCeilSCU = Union[int, float, bool, Real, SupportsCeil]
SupportsCeilSCT = (int, float, bool, Real, SupportsCeil)
assert SupportsCeilSCU.__args__ == SupportsCeilSCT  # type: ignore


@runtime_checkable
class SupportsDivmod(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __divmod__(self, other: Any) -> Tuple[_T_co, _T_co]:
        pass

    @abstractmethod
    def __rdivmod__(self, other: Any) -> Tuple[_T_co, _T_co]:
        pass


_assert_isinstance(int, bool, float, Decimal, Fraction, target_t=SupportsDivmod)
SupportsDivmodSCU = Union[int, float, bool, Real, SupportsDivmod]
SupportsDivmodSCT = (int, float, bool, Real, SupportsDivmod)
assert SupportsDivmodSCU.__args__ == SupportsDivmodSCT  # type: ignore


@runtime_checkable
class SupportsNumeratorDenominatorProperties(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @property
    def numerator(self) -> int:
        pass

    @property
    def denominator(self) -> int:
        pass


_assert_isinstance(int, bool, Fraction, target_t=SupportsNumeratorDenominatorProperties)
SupportsNumeratorDenominatorPropertiesSCU = Union[
    int,
    bool,
    Rational,
    SupportsNumeratorDenominatorProperties,
]
SupportsNumeratorDenominatorPropertiesSCT = (
    int,
    bool,
    Rational,
    SupportsNumeratorDenominatorProperties,
)
assert SupportsNumeratorDenominatorPropertiesSCU.__args__ == SupportsNumeratorDenominatorPropertiesSCT  # type: ignore


@runtime_checkable
class SupportsNumeratorDenominatorMethods(
    Protocol,
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def numerator(self) -> SupportsInt:
        pass

    @abstractmethod
    def denominator(self) -> SupportsInt:
        pass


SupportsNumeratorDenominatorMixedU = Union[
    SupportsNumeratorDenominatorProperties,
    SupportsNumeratorDenominatorMethods,
]
SupportsNumeratorDenominatorMixedT = (
    SupportsNumeratorDenominatorProperties,
    SupportsNumeratorDenominatorMethods,
)
assert SupportsNumeratorDenominatorMixedU.__args__ == SupportsNumeratorDenominatorMixedT  # type: ignore

SupportsNumeratorDenominatorMixedSCU = Union[
    SupportsNumeratorDenominatorPropertiesSCU,
    SupportsNumeratorDenominatorMethods,
]
SupportsNumeratorDenominatorMixedSCT = SupportsNumeratorDenominatorPropertiesSCT + (
    SupportsNumeratorDenominatorMethods,
)
assert SupportsNumeratorDenominatorMixedSCU.__args__ == SupportsNumeratorDenominatorMixedSCT  # type: ignore


@runtime_checkable
class SupportsComplexOps(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __add__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __radd__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __sub__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rsub__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __mul__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rmul__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __truediv__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rtruediv__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __neg__(self) -> _T_co:
        pass

    @abstractmethod
    def __pos__(self) -> _T_co:
        pass


_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsComplexOps)
SupportsComplexOpsSCU = Union[int, float, bool, Complex, SupportsComplexOps]
SupportsComplexOpsSCT = (int, float, bool, Complex, SupportsComplexOps)
assert SupportsComplexOpsSCU.__args__ == SupportsComplexOpsSCT  # type: ignore


@runtime_checkable
class SupportsComplexPow(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __pow__(self, exponent: Any) -> _T_co:
        pass

    @abstractmethod
    def __rpow__(self, exponent: Any) -> _T_co:
        pass


_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsComplexPow)
SupportsComplexPowSCU = Union[int, float, bool, Complex, SupportsComplexPow]
SupportsComplexPowSCT = (int, float, bool, Complex, SupportsComplexPow)
assert SupportsComplexPowSCU.__args__ == SupportsComplexPowSCT  # type: ignore


@runtime_checkable
class SupportsRealOps(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __lt__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __le__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __ge__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __gt__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __floordiv__(self, other: Any) -> Any:  # TODO(posita): should be int
        pass

    @abstractmethod
    def __rfloordiv__(self, other: Any) -> Any:  # TODO(posita): should be int
        pass

    @abstractmethod
    def __mod__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rmod__(self, other: Any) -> _T_co:
        pass


_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=SupportsRealOps)
SupportsRealOpsSCU = Union[int, float, bool, Real, SupportsRealOps]
SupportsRealOpsSCT = (int, float, bool, Real, SupportsRealOps)
assert SupportsRealOpsSCU.__args__ == SupportsRealOpsSCT  # type: ignore


@runtime_checkable
class SupportsIntegralOps(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __lshift__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rlshift__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rshift__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rrshift__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __and__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rand__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __xor__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __rxor__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __or__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __ror__(self, other: Any) -> _T_co:
        pass

    @abstractmethod
    def __invert__(self) -> _T_co:
        pass


_assert_isinstance(int, bool, target_t=SupportsIntegralOps)
SupportsIntegralOpsSCU = Union[int, bool, Integral, SupportsIntegralOps]
SupportsIntegralOpsSCT = (int, bool, Integral, SupportsIntegralOps)
assert SupportsIntegralOpsSCU.__args__ == SupportsIntegralOpsSCT  # type: ignore


@runtime_checkable
class SupportsIntegralPow(
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    @abstractmethod
    def __pow__(self, exponent: Any, modulus: Optional[Any] = None) -> _T_co:
        pass

    @abstractmethod
    def __rpow__(self, exponent: Any, modulus: Optional[Any] = None) -> _T_co:
        pass


_assert_isinstance(int, bool, target_t=SupportsIntegralPow)
SupportsIntegralPowSCU = Union[int, bool, Integral, SupportsIntegralPow]
SupportsIntegralPowSCT = (int, bool, Integral, SupportsIntegralPow)
assert SupportsIntegralPowSCU.__args__ == SupportsIntegralPowSCT  # type: ignore


@runtime_checkable
class RealLike(
    SupportsAbs[_T_co],
    SupportsFloat,
    SupportsRealOps[_T_co],
    SupportsComplexOps[_T_co],
    SupportsComplexPow[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    # Must be able to instantiate it
    @abstractmethod
    def __init__(self, *args: Any, **kw: Any):
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass


_assert_isinstance(int, float, bool, Decimal, Fraction, target_t=RealLike)
RealLikeSCU = Union[int, float, bool, Real, RealLike]
RealLikeSCT = (int, float, bool, Real, RealLike)
assert RealLikeSCU.__args__ == RealLikeSCT  # type: ignore


@runtime_checkable
class RationalLikeProperties(
    SupportsAbs[_T_co],
    SupportsFloat,
    SupportsNumeratorDenominatorProperties,
    SupportsRealOps[_T_co],
    SupportsComplexOps[_T_co],
    SupportsComplexPow[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    # Must be able to instantiate it
    @abstractmethod
    def __init__(self, *args: Any, **kw: Any):
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass


_assert_isinstance(int, bool, Fraction, target_t=RationalLikeProperties)
RationalLikePropertiesSCU = Union[
    int,
    bool,
    Rational,
    RationalLikeProperties,
]
RationalLikePropertiesSCT = (
    int,
    bool,
    Rational,
    RationalLikeProperties,
)
assert RationalLikePropertiesSCU.__args__ == RationalLikePropertiesSCT  # type: ignore


@runtime_checkable
class RationalLikeMethods(
    SupportsAbs[_T_co],
    SupportsFloat,
    SupportsNumeratorDenominatorMethods,
    SupportsRealOps[_T_co],
    SupportsComplexOps[_T_co],
    SupportsComplexPow[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    # Must be able to instantiate it
    @abstractmethod
    def __init__(self, *args: Any, **kw: Any):
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass


RationalLikeMixedU = Union[RationalLikeProperties, RationalLikeMethods]
RationalLikeMixedT = (RationalLikeProperties, RationalLikeMethods)
RationalLikeMixedSCU = Union[
    RationalLikePropertiesSCU,
    RationalLikeMethods,
]
RationalLikeMixedSCT = RationalLikePropertiesSCT + (RationalLikeMethods,)
assert RationalLikeMixedSCU.__args__ == RationalLikeMixedSCT  # type: ignore


@runtime_checkable
class IntegralLike(
    SupportsAbs[_T_co],
    SupportsFloat,
    SupportsIndex,
    SupportsInt,
    SupportsIntegralOps[_T_co],
    SupportsIntegralPow[_T_co],
    SupportsRealOps[_T_co],
    SupportsComplexOps[_T_co],
    Protocol[_T_co],
    metaclass=CachingProtocolMeta,
):
    r"""
    TODO(posita): Document this!
    """
    __slots__: Union[str, Iterable[str]] = ()

    # Must be able to instantiate it
    @abstractmethod
    def __init__(self, *args: Any, **kw: Any):
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass


_assert_isinstance(int, bool, target_t=IntegralLike)
IntegralLikeSCU = Union[int, bool, Integral, IntegralLike]
IntegralLikeSCT = (int, bool, Integral, IntegralLike)
assert IntegralLikeSCU.__args__ == IntegralLikeSCT  # type: ignore


# ---- Functions -----------------------------------------------------------------------


@beartype
def ceil(operand: SupportsCeil):
    r"""
    TODO(posita): Document this!
    """
    assert isinstance(operand, SupportsFloat)

    return math.ceil(operand)


@beartype
def floor(operand: SupportsFloor):
    r"""
    TODO(posita): Document this!
    """
    assert isinstance(operand, SupportsFloat)

    return math.floor(operand)


@beartype
def trunc(operand: SupportsTrunc):
    r"""
    TODO(posita): Document this!
    """
    assert isinstance(operand, SupportsFloat)

    return math.trunc(operand)


@beartype
def numerator(operand: SupportsNumeratorDenominatorMixedU):
    r"""
    TODO(posita): Document this!
    """
    if hasattr(operand, "numerator"):
        if callable(operand.numerator):
            return operand.numerator()
        else:
            return operand.numerator
    else:
        raise TypeError(f"{operand!r} has no numerator")


@beartype
def denominator(operand: SupportsNumeratorDenominatorMixedU):
    r"""
    TODO(posita): Document this!
    """
    if hasattr(operand, "denominator"):
        if callable(operand.denominator):
            return operand.denominator()
        else:
            return operand.denominator
    else:
        raise TypeError(f"{operand!r} has no denominator")



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import SupportsTrunc, SupportsTruncSCT, SupportsTruncSCU, trunc

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def func(arg: SupportsTrunc):
    assert isinstance(arg, SupportsTrunc), f"{arg!r}"


@beartype
def func_t(arg: SupportsTruncSCU):
    assert isinstance(arg, SupportsTruncSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_trunc() -> None:
    bool_val: SupportsTrunc = True
    int_val: SupportsTrunc = -273
    float_val: SupportsTrunc = -273.15
    frac_val: SupportsTrunc = Fraction(-27315, 100)
    dec_val: SupportsTrunc = Decimal("-273.15")
    nw_val: SupportsTrunc = Numberwang(-273)
    nwd_val: SupportsTrunc = NumberwangDerived(-273)
    nwr_val: SupportsTrunc = NumberwangRegistered(-273)
    wn_val: SupportsTrunc = Wangernumb(-273.15)
    wnd_val: SupportsTrunc = WangernumbDerived(-273.15)
    wnr_val: SupportsTrunc = WangernumbRegistered(-273.15)

    for good_val in (
        bool_val,
        int_val,
        float_val,
        frac_val,
        dec_val,
        nw_val,
        nwd_val,
        nwr_val,
        wn_val,
        wnd_val,
        wnr_val,
    ):
        assert isinstance(good_val, SupportsTrunc), f"{good_val!r}"
        assert isinstance(good_val, SupportsTruncSCT), f"{good_val!r}"
        assert trunc(good_val), f"{good_val!r}"

    for bad_val in (
        complex(-273.15),
        "-273.15",
    ):
        assert not isinstance(bad_val, SupportsTrunc), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsTruncSCT), f"{bad_val!r}"


def test_trunc_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        -273.15,
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
    ):
        func(cast(SupportsTrunc, good_val))
        func_t(cast(SupportsTruncSCU, good_val))

    for bad_val in (
        complex(-273.15),
        "-273.15",
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsTrunc, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsTruncSCU, bad_val))


def test_trunc_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsTrunc = numpy.uint8(2)
    uint16_val: SupportsTrunc = numpy.uint16(273)
    uint32_val: SupportsTrunc = numpy.uint32(273)
    uint64_val: SupportsTrunc = numpy.uint64(273)
    int8_val: SupportsTrunc = numpy.int8(-2)
    int16_val: SupportsTrunc = numpy.int16(-273)
    int32_val: SupportsTrunc = numpy.int32(-273)
    int64_val: SupportsTrunc = numpy.int64(-273)
    float16_val: SupportsTrunc = numpy.float16(-1.8)
    float32_val: SupportsTrunc = numpy.float32(-273.15)
    float64_val: SupportsTrunc = numpy.float64(-273.15)
    float128_val: SupportsTrunc = numpy.float128(-273.15)

    # numpy.float64 seems to have a closer relationship to the native float than the
    # other numpy.float* types
    for good_val in (float64_val,):
        assert isinstance(good_val, SupportsTrunc), f"{good_val!r}"
        assert isinstance(good_val, SupportsTruncSCT), f"{good_val!r}"
        assert trunc(good_val), f"{good_val!r}"

    for lying_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
        float16_val,
        float32_val,
        float128_val,
    ):
        # The pure protocol approach catches this
        assert not isinstance(lying_val, SupportsTrunc), f"{lying_val!r}"

        # The short-circuiting approach does not
        assert isinstance(lying_val, SupportsTruncSCT), f"{lying_val!r}"

    for bad_val in (
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert not isinstance(bad_val, SupportsTrunc), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsTruncSCT), f"{bad_val!r}"


def test_trunc_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    # numpy.float64 seems to have a closer relationship to the native float than the
    # other numpy.float* types
    for good_val in (numpy.float64(-273.15),):
        func(cast(SupportsTrunc, good_val))
        func_t(cast(SupportsTruncSCU, good_val))

    for lying_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float128(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsTrunc, lying_val))

        func_t(cast(SupportsTruncSCU, lying_val))

    for bad_val in (
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsTrunc, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsTruncSCU, bad_val))


def test_trunc_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    integer_val: SupportsTrunc = sympy.Integer(-273)
    rational_val: SupportsTrunc = sympy.Rational(-27315, 100)
    float_val: SupportsTrunc = sympy.Float(-273.15)
    sym_val: SupportsTrunc = sympy.symbols("x")

    for good_val in (
        integer_val,
        rational_val,
        float_val,
    ):
        assert isinstance(good_val, SupportsTrunc), f"{good_val!r}"
        assert isinstance(good_val, SupportsTruncSCT), f"{good_val!r}"
        assert trunc(good_val), f"{good_val!r}"

    for lying_val in (sym_val,):
        assert isinstance(lying_val, SupportsTrunc), f"{lying_val!r}"
        assert isinstance(lying_val, SupportsTruncSCT), f"{lying_val!r}"

        # Relationals have, but don't implement this function
        # TODO(posita): Can we fix this?
        with pytest.raises(TypeError):
            trunc(lying_val)


def test_trunc_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Integer(-273),
        sympy.Rational(-27315, 100),
        sympy.Float(-273.15),
    ):
        func(cast(SupportsTrunc, good_val))
        func_t(cast(SupportsTruncSCU, good_val))

    for lying_val in (sympy.symbols("x"),):
        func(cast(SupportsTrunc, good_val))
        func_t(cast(SupportsTruncSCU, good_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from typing import Tuple

import pytest

from numerary import IntegralLikeSCU, RealLikeSCU
from numerary.bt import beartype

__all__ = ()


# ---- Tests ---------------------------------------------------------------------------


def test_beartype_detection() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    @beartype
    def _real_like_identity(arg: RealLikeSCU) -> RealLikeSCU:
        return arg

    with pytest.raises(roar.BeartypeException):
        _real_like_identity("-273")  # type: ignore

    @beartype
    def _lies_all_lies(arg: RealLikeSCU) -> Tuple[str]:
        return (arg,)  # type: ignore

    with pytest.raises(roar.BeartypeException):
        _lies_all_lies(-273)


def test_beartype_validators() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")
    from beartype.vale import Is

    from numerary.types import Annotated

    NonZero = Annotated[IntegralLikeSCU, Is[lambda x: x != 0]]

    @beartype
    def _divide_it(n: IntegralLikeSCU, d: NonZero) -> RealLikeSCU:
        return n / d

    with pytest.raises(roar.BeartypeException):
        _divide_it(0, 0)

    with pytest.raises(roar.BeartypeException):
        _divide_it(0, "1")  # type: ignore

    If = Annotated[
        Tuple[str, ...],
        Is[
            lambda x: x
            == (
                "If you can dream—and not make dreams your master;",
                "If you can think—and not make thoughts your aim;",
            )
        ],
    ]

    @beartype
    def _if(lines: If) -> Tuple[str, ...]:
        return (
            "If you can meet with Triumph and Disaster",
            "And treat those two impostors just the same;",
        )

    with pytest.raises(roar.BeartypeException):
        _if(())



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import math
from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import (
    RationalLikeMixedSCT,
    RationalLikeMixedSCU,
    RationalLikeMixedT,
    RationalLikeMixedU,
    RationalLikeProperties,
)

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def func(arg: RationalLikeMixedU):
    assert isinstance(arg, RationalLikeMixedT), f"{arg!r}"


@beartype
def func_t(arg: RationalLikeMixedSCU):
    assert isinstance(arg, RationalLikeMixedSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_rational_like() -> None:
    bool_val: RationalLikeProperties = True
    int_val: RationalLikeProperties = -273
    frac_val: RationalLikeProperties = Fraction(-27315, 100)
    nw_val: RationalLikeProperties = Numberwang(-273)
    nwd_val: RationalLikeProperties = NumberwangDerived(-273)
    nwr_val: RationalLikeProperties = NumberwangRegistered(-273)

    for good_val in (
        bool_val,
        int_val,
        frac_val,
        nw_val,
        nwd_val,
        nwr_val,
    ):
        assert isinstance(good_val, RationalLikeMixedT), f"{good_val!r}"
        assert isinstance(good_val, RationalLikeMixedSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"
        assert good_val - 1 < good_val, f"{good_val!r}"
        assert good_val - 1 <= good_val, f"{good_val!r}"
        assert good_val >= good_val - 1, f"{good_val!r}"
        assert good_val > good_val - 1, f"{good_val!r}"
        assert math.trunc(good_val), f"{good_val!r}"
        assert math.floor(good_val), f"{good_val!r}"
        assert math.ceil(good_val), f"{good_val!r}"
        assert good_val.numerator, f"{good_val!r}"
        assert good_val.denominator, f"{good_val!r}"

    for bad_val in (
        -273.15,
        complex(-273.15),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        assert not isinstance(bad_val, RationalLikeMixedT), f"{bad_val!r}"
        assert not isinstance(bad_val, RationalLikeMixedSCT), f"{bad_val!r}"


def test_rational_like_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        Fraction(-27315, 100),
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
    ):
        func(cast(RationalLikeMixedU, good_val))
        func_t(cast(RationalLikeMixedSCU, good_val))

    for bad_val in (
        -273.15,
        complex(-273.15),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(RationalLikeMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(RationalLikeMixedSCU, bad_val))


def test_rational_like_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: RationalLikeProperties = numpy.uint8(2)
    uint16_val: RationalLikeProperties = numpy.uint16(273)
    uint32_val: RationalLikeProperties = numpy.uint32(273)
    uint64_val: RationalLikeProperties = numpy.uint64(273)
    int8_val: RationalLikeProperties = numpy.int8(-2)
    int16_val: RationalLikeProperties = numpy.int16(-273)
    int32_val: RationalLikeProperties = numpy.int32(-273)
    int64_val: RationalLikeProperties = numpy.int64(-273)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
    ):
        assert isinstance(good_val, RationalLikeMixedT), f"{good_val!r}"
        assert isinstance(good_val, RationalLikeMixedSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"
        assert good_val - 1 < good_val, f"{good_val!r}"
        assert good_val - 1 <= good_val, f"{good_val!r}"
        assert good_val >= good_val - 1, f"{good_val!r}"
        assert good_val > good_val - 1, f"{good_val!r}"
        assert good_val.numerator, f"{good_val!r}"
        assert good_val.denominator, f"{good_val!r}"

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert not isinstance(bad_val, RationalLikeMixedT), f"{bad_val!r}"
        assert not isinstance(bad_val, RationalLikeMixedSCT), f"{bad_val!r}"


def test_rational_like_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
    ):
        func(cast(RationalLikeMixedU, good_val))
        func_t(cast(RationalLikeMixedSCU, good_val))

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(RationalLikeMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(RationalLikeMixedSCU, bad_val))


def test_rational_like_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    integral_val: RationalLikeProperties = sympy.Integer(-273)
    rational_val: RationalLikeProperties = sympy.Rational(-27315, 100)

    for good_val in (
        integral_val,
        rational_val,
    ):
        assert isinstance(good_val, RationalLikeMixedT), f"{good_val!r}"
        assert isinstance(good_val, RationalLikeMixedSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"
        assert good_val - 1 < good_val, f"{good_val!r}"
        assert good_val - 1 <= good_val, f"{good_val!r}"
        assert good_val >= good_val - 1, f"{good_val!r}"
        assert good_val > good_val - 1, f"{good_val!r}"
        assert math.trunc(good_val), f"{good_val!r}"
        assert math.floor(good_val), f"{good_val!r}"
        assert math.ceil(good_val), f"{good_val!r}"
        assert good_val.numerator, f"{good_val!r}"
        assert good_val.denominator, f"{good_val!r}"

    for bad_val in (
        sympy.Float(-273.15),
        sympy.symbols("x"),
    ):
        assert not isinstance(bad_val, RationalLikeMixedT), f"{bad_val!r}"
        assert not isinstance(bad_val, RationalLikeMixedSCT), f"{bad_val!r}"


def test_rational_like_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Rational(-27315, 100),
        sympy.Integer(-273),
    ):
        func(cast(RationalLikeMixedU, good_val))
        func_t(cast(RationalLikeMixedSCU, good_val))

    for bad_val in (
        sympy.Float(-273.15),
        sympy.symbols("x"),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(RationalLikeMixedU, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(RationalLikeMixedSCU, bad_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import (
    SupportsComplexOps,
    SupportsComplexOpsSCT,
    SupportsComplexOpsSCU,
    SupportsComplexPow,
    SupportsComplexPowSCT,
    SupportsComplexPowSCU,
)

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def ops_func(arg: SupportsComplexOps):
    assert isinstance(arg, SupportsComplexOps), f"{arg!r}"


@beartype
def ops_func_t(arg: SupportsComplexOpsSCU):
    assert isinstance(arg, SupportsComplexOpsSCT), f"{arg!r}"


@beartype
def pow_func(arg: SupportsComplexPow):
    assert isinstance(arg, SupportsComplexPow), f"{arg!r}"


@beartype
def pow_func_t(arg: SupportsComplexPowSCU):
    assert isinstance(arg, SupportsComplexPowSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_supports_complex_ops_pow() -> None:
    bool_val: SupportsComplexOps = True
    int_val: SupportsComplexOps = -273
    float_val: SupportsComplexOps = -273.15
    complex_val: SupportsComplexOps = complex(-273.15)
    frac_val: SupportsComplexOps = Fraction(-27315, 100)
    dec_val: SupportsComplexOps = Decimal("-273.15")
    nw_val: SupportsComplexOps = Numberwang(-273)
    nwd_val: SupportsComplexOps = NumberwangDerived(-273)
    nwr_val: SupportsComplexOps = NumberwangRegistered(-273)
    wn_val: SupportsComplexOps = Wangernumb(-273.15)
    wnd_val: SupportsComplexOps = WangernumbDerived(-273.15)
    wnr_val: SupportsComplexOps = WangernumbRegistered(-273.15)
    _: SupportsComplexPow
    _ = True
    _ = -273
    _ = -273.15
    _ = complex(-273.15)
    _ = Fraction(-27315, 100)
    _ = Decimal("-273.15")
    _ = Numberwang(-273)
    _ = NumberwangDerived(-273)
    _ = NumberwangRegistered(-273)
    _ = Wangernumb(-273.15)
    _ = WangernumbDerived(-273.15)
    _ = WangernumbRegistered(-273.15)

    for good_val in (
        bool_val,
        int_val,
        float_val,
        complex_val,
        frac_val,
        dec_val,
        nw_val,
        nwd_val,
        nwr_val,
        wn_val,
        wnd_val,
        wnr_val,
    ):
        assert isinstance(good_val, SupportsComplexOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexOpsSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"

    for bad_val in ("-273.15",):
        assert not isinstance(bad_val, SupportsComplexOps), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsComplexOpsSCT), f"{bad_val!r}"


def test_supports_complex_ops_pow_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        -273.15,
        complex(-273.15),
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
    ):
        ops_func(cast(SupportsComplexOps, good_val))
        ops_func_t(cast(SupportsComplexOpsSCU, good_val))
        pow_func(cast(SupportsComplexPow, good_val))
        pow_func_t(cast(SupportsComplexPowSCU, good_val))

    for bad_val in ("-273.15",):
        with pytest.raises(roar.BeartypeException):
            ops_func(cast(SupportsComplexOps, bad_val))

        with pytest.raises(roar.BeartypeException):
            ops_func_t(cast(SupportsComplexOpsSCU, bad_val))

        with pytest.raises(roar.BeartypeException):
            pow_func(cast(SupportsComplexPow, bad_val))

        with pytest.raises(roar.BeartypeException):
            pow_func_t(cast(SupportsComplexPowSCU, bad_val))


def test_supports_complex_ops_pow_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsComplexOps = numpy.uint8(2)
    uint16_val: SupportsComplexOps = numpy.uint16(273)
    uint32_val: SupportsComplexOps = numpy.uint32(273)
    uint64_val: SupportsComplexOps = numpy.uint64(273)
    int8_val: SupportsComplexOps = numpy.int8(-2)
    int16_val: SupportsComplexOps = numpy.int16(-273)
    int32_val: SupportsComplexOps = numpy.int32(-273)
    int64_val: SupportsComplexOps = numpy.int64(-273)
    float16_val: SupportsComplexOps = numpy.float16(-1.8)
    float32_val: SupportsComplexOps = numpy.float32(-273.15)
    float64_val: SupportsComplexOps = numpy.float64(-273.15)
    float128_val: SupportsComplexOps = numpy.float128(-273.15)
    csingle_val: SupportsComplexOps = numpy.float32(-273.15)
    cdouble_val: SupportsComplexOps = numpy.float64(-273.15)
    clongdouble_val: SupportsComplexOps = numpy.float128(-273.15)
    _: SupportsComplexPow
    _ = numpy.uint8(2)
    _ = numpy.uint16(273)
    _ = numpy.uint32(273)
    _ = numpy.uint64(273)
    _ = numpy.int8(-2)
    _ = numpy.int16(-273)
    _ = numpy.int32(-273)
    _ = numpy.int64(-273)
    _ = numpy.float16(-1.8)
    _ = numpy.float32(-273.15)
    _ = numpy.float64(-273.15)
    _ = numpy.float128(-273.15)
    _ = numpy.float32(-273.15)
    _ = numpy.float64(-273.15)
    _ = numpy.float128(-273.15)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
        float16_val,
        float32_val,
        float64_val,
        float128_val,
        csingle_val,
        cdouble_val,
        clongdouble_val,
    ):
        assert isinstance(good_val, SupportsComplexOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexOpsSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"

        # numpy.uint*.__neg__ means something different than what we're testing for
        if not isinstance(
            good_val,
            (numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64),
        ):
            assert 0 - good_val == -good_val, f"{good_val!r}"

        assert isinstance(good_val, SupportsComplexPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"


def test_supports_complex_ops_pow_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        ops_func(cast(SupportsComplexOps, good_val))
        ops_func_t(cast(SupportsComplexOpsSCU, good_val))
        pow_func(cast(SupportsComplexPow, good_val))
        pow_func_t(cast(SupportsComplexPowSCU, good_val))


def test_supports_complex_ops_pow_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    integer_val: SupportsComplexOps = sympy.Integer(-273)
    rational_val: SupportsComplexOps = sympy.Rational(-27315, 100)
    float_val: SupportsComplexOps = sympy.Float(-273.15)
    # sym_val: SupportsComplexOps = (sympy.symbols("x"),)
    _: SupportsComplexPow
    _ = sympy.Integer(-273)
    _ = sympy.Rational(-27315, 100)
    _ = sympy.Float(-273.15)
    # _ = (sympy.symbols("x"),)

    for good_val in (
        integer_val,
        rational_val,
        float_val,
        # TODO(posita): Can we fix this?
        # sym_val,
    ):
        assert isinstance(good_val, SupportsComplexOps), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexOpsSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"

        assert isinstance(good_val, SupportsComplexPow), f"{good_val!r}"
        assert isinstance(good_val, SupportsComplexPowSCT), f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"


def test_supports_complex_ops_pow_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Integer(-273),
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
        sympy.symbols("x"),
    ):
        ops_func(cast(SupportsComplexOps, good_val))
        ops_func_t(cast(SupportsComplexOpsSCU, good_val))
        pow_func(cast(SupportsComplexPow, good_val))
        pow_func_t(cast(SupportsComplexPowSCU, good_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import math
import operator
from typing import Callable, Tuple, TypeVar

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Types ---------------------------------------------------------------------------


_T_co = TypeVar("_T_co", covariant=True)
_UnaryOperatorT = Callable[[_T_co], _T_co]
_BinaryOperatorT = Callable[[_T_co, _T_co], _T_co]


# ---- Data ----------------------------------------------------------------------------


BINOPS_REAL: Tuple[_BinaryOperatorT, ...] = (
    operator.__add__,
    operator.__eq__,
    operator.__floordiv__,
    operator.__ge__,
    operator.__gt__,
    operator.__le__,
    operator.__lt__,
    operator.__mod__,
    operator.__mul__,
    operator.__ne__,
    operator.__pow__,
    operator.__sub__,
    operator.__truediv__,
)

BINOPS_INTEGRAL: Tuple[_BinaryOperatorT, ...] = (
    operator.__and__,
    operator.__lshift__,
    operator.__or__,
    operator.__rshift__,
    operator.__xor__,
)

UNOPS_REAL: Tuple[_UnaryOperatorT, ...] = (
    int,
    round,
    math.ceil,
    math.floor,
    math.trunc,
    operator.__abs__,
    operator.__neg__,
    operator.__pos__,
)

UNOPS_REAL_DERIVED: Tuple[_UnaryOperatorT, ...] = UNOPS_REAL + (
    complex,
    float,
)

UNOPS_INTEGRAL: Tuple[_UnaryOperatorT, ...] = (operator.__invert__,)

UNOPS_INTEGRAL_DERIVED: Tuple[_UnaryOperatorT, ...] = UNOPS_INTEGRAL + (
    operator.__index__,
)

# ---- Tests ---------------------------------------------------------------------------


def test_numberwang_binops() -> None:
    for binop in BINOPS_REAL + BINOPS_INTEGRAL:
        assert binop(Numberwang(-273), Numberwang(42)) == binop(
            -273, 42
        ), f"op: {binop}"


def test_numberwang_unops() -> None:
    for unop in UNOPS_REAL + UNOPS_INTEGRAL:
        assert unop(Numberwang(-273)) == unop(-273), f"op: {unop}"


def test_numberwang_registered_binops() -> None:
    for binop in BINOPS_REAL + BINOPS_INTEGRAL:
        assert binop(NumberwangRegistered(-273), NumberwangRegistered(42)) == binop(
            -273, 42
        ), f"op: {binop}"


def test_numberwang_registered_unops() -> None:
    for unop in UNOPS_REAL + UNOPS_INTEGRAL:
        assert unop(NumberwangRegistered(-273)) == unop(-273), f"op: {unop}"


def test_numberwang_derived_binops() -> None:
    for binop in BINOPS_REAL + BINOPS_INTEGRAL:
        assert binop(NumberwangDerived(-273), NumberwangDerived(42)) == binop(
            -273, 42
        ), f"op: {binop}"


def test_numberwang_derived_unops() -> None:
    for unop in UNOPS_REAL_DERIVED + UNOPS_INTEGRAL_DERIVED:
        assert unop(NumberwangDerived(-273)) == unop(-273), f"op: {unop}"


def test_wangernum_binops() -> None:
    for binop in BINOPS_REAL:
        assert binop(Wangernumb(-273.15), Wangernumb(1.618)) == binop(
            -273.15, 1.618
        ), f"op: {binop}"


def test_wangernum_unops() -> None:
    for unop in UNOPS_REAL:
        assert unop(Wangernumb(-273.15)) == unop(-273.15), f"op: {unop}"


def test_wangernum_registered_binops() -> None:
    for binop in BINOPS_REAL:
        assert binop(
            WangernumbRegistered(-273.15), WangernumbRegistered(1.618)
        ) == binop(-273.15, 1.618), f"op: {binop}"


def test_wangernum_registered_unops() -> None:
    for unop in UNOPS_REAL:
        assert unop(WangernumbRegistered(-273.15)) == unop(-273.15), f"op: {unop}"


def test_wangernum_derived_binops() -> None:
    for binop in BINOPS_REAL:
        assert binop(WangernumbDerived(-273.15), WangernumbDerived(1.618)) == binop(
            -273.15, 1.618
        ), f"op: {binop}"


def test_wangernum_derived_unops() -> None:
    for unop in UNOPS_REAL_DERIVED:
        assert unop(WangernumbDerived(-273.15)) == unop(-273.15), f"op: {unop}"



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from typing import Tuple

__all__ = ()


# ---- Data ----------------------------------------------------------------------------


__version__: Tuple[int, int, int] = (0, 0, 4)
__vers_str__ = ".".join(str(_) for _ in __version__)
__release__ = "v" + __vers_str__



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import SupportsDivmod, SupportsDivmodSCT, SupportsDivmodSCU

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def func(arg: SupportsDivmod):
    assert isinstance(arg, SupportsDivmod), f"{arg!r}"


@beartype
def func_t(arg: SupportsDivmodSCU):
    assert isinstance(arg, SupportsDivmodSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_supports_divmod() -> None:
    bool_val: SupportsDivmod = True
    int_val: SupportsDivmod = -273
    float_val: SupportsDivmod = -273.15
    frac_val: SupportsDivmod = Fraction(-27315, 100)
    dec_val: SupportsDivmod = Decimal("-273.15")
    # These have inherited this interface by deriving from number tower ABCs
    nwd_val: SupportsDivmod = NumberwangDerived(-273)
    wnd_val: SupportsDivmod = WangernumbDerived(-273.15)

    for good_val in (
        bool_val,
        int_val,
        float_val,
        frac_val,
        dec_val,
        nwd_val,
        wnd_val,
    ):
        assert isinstance(good_val, SupportsDivmod), f"{good_val!r}"
        assert isinstance(good_val, SupportsDivmodSCT), f"{good_val!r}"
        assert divmod(good_val, good_val), f"{good_val!r}"

    for lying_val in (
        # These have lied about supporting this interface when they registered
        # themselves in the number tower
        NumberwangRegistered(-273),
        WangernumbRegistered(-273.15),
    ):
        # The pure protocol approach catches this
        assert not isinstance(lying_val, SupportsDivmod), f"{lying_val!r}"

        # The short-circuiting approach does not
        assert isinstance(lying_val, SupportsDivmodSCT), f"{lying_val!r}"

    for bad_val in (
        # TODO(posita): fix this
        # complex(-273.15),
        Numberwang(-273),
        Wangernumb(-273.15),
        "-273.15",
    ):
        assert not isinstance(bad_val, SupportsDivmod), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsDivmodSCT), f"{bad_val!r}"


def test_supports_divmod_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        -273.15,
        Fraction(-27315, 100),
        Decimal("-273.15"),
        # These have inherited this interface by deriving from number tower ABCs
        NumberwangDerived(-273),
        WangernumbDerived(-273.15),
    ):
        func(cast(SupportsDivmod, good_val))
        func_t(cast(SupportsDivmodSCU, good_val))

    for lying_val in (
        # These have lied about supporting this interface when they registered
        # themselves in the number tower
        NumberwangRegistered(-273),
        WangernumbRegistered(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsDivmod, lying_val))

        func_t(cast(SupportsDivmodSCU, lying_val))

    for bad_val in (
        # TODO(posita): fix this
        # complex(-273.15),
        Numberwang(-273),
        Wangernumb(-273.15),
        "-273.15",
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsDivmod, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsDivmodSCU, bad_val))


def test_supports_divmod_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsDivmod = numpy.uint8(2)
    uint16_val: SupportsDivmod = numpy.uint16(273)
    uint32_val: SupportsDivmod = numpy.uint32(273)
    uint64_val: SupportsDivmod = numpy.uint64(273)
    int8_val: SupportsDivmod = numpy.int8(-2)
    int16_val: SupportsDivmod = numpy.int16(-273)
    int32_val: SupportsDivmod = numpy.int32(-273)
    int64_val: SupportsDivmod = numpy.int64(-273)
    float16_val: SupportsDivmod = numpy.float16(-1.8)
    float32_val: SupportsDivmod = numpy.float32(-273.15)
    float64_val: SupportsDivmod = numpy.float64(-273.15)
    float128_val: SupportsDivmod = numpy.float128(-273.15)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
        float16_val,
        float32_val,
        float64_val,
        float128_val,
    ):
        assert isinstance(good_val, SupportsDivmod), f"{good_val!r}"
        assert isinstance(good_val, SupportsDivmodSCT), f"{good_val!r}"
        assert divmod(good_val, good_val), f"{good_val!r}"

    for bad_val in (
        # TODO(posita): fix these
        # numpy.csingle(-273.15),
        # numpy.cdouble(-273.15),
        # numpy.clongdouble(-273.15),
        "-273.15",  # TODO(posita): remove me
    ):
        assert not isinstance(bad_val, SupportsDivmod), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsDivmodSCT), f"{bad_val!r}"


def test_supports_divmod_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
    ):
        func(cast(SupportsDivmod, good_val))
        func_t(cast(SupportsDivmodSCU, good_val))

    for bad_val in (
        # TODO(posita): fix these
        # numpy.csingle(-273.15),
        # numpy.cdouble(-273.15),
        # numpy.clongdouble(-273.15),
        "-273.15",  # TODO(posita): remove me
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(SupportsDivmod, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(SupportsDivmodSCU, bad_val))


def test_supports_divmod_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    integer_val: SupportsDivmod = sympy.Integer(-273)
    rational_val: SupportsDivmod = sympy.Rational(-27315, 100)
    float_val: SupportsDivmod = sympy.Float(-273.15)
    # sym_val: SupportsDivmod = (sympy.symbols("x"),)

    for good_val in (
        integer_val,
        rational_val,
        float_val,
        # TODO(posita): Can we fix this?
        # sym_val,
    ):
        assert isinstance(good_val, SupportsDivmod), f"{good_val!r}"
        assert isinstance(good_val, SupportsDivmodSCT), f"{good_val!r}"

        # Symbolic relationals can't be reduced to a boolean
        if isinstance(good_val, sympy.core.symbol.Symbol):
            assert divmod(good_val, good_val), f"{good_val!r}"
        else:
            assert divmod(good_val, good_val), f"{good_val!r}"


def test_supports_divmod_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Integer(-273),
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
        sympy.symbols("x"),
    ):
        func(cast(SupportsDivmod, good_val))
        func_t(cast(SupportsDivmodSCU, good_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

from .types import *  # noqa: F401,F403
from .version import __version__  # noqa: F401

__all__ = ()



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import sys
from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary.bt import beartype
from numerary.types import (
    SupportsCeil,
    SupportsCeilSCT,
    SupportsCeilSCU,
    SupportsFloor,
    SupportsFloorSCT,
    SupportsFloorSCU,
    ceil,
    floor,
)

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def ceil_func(arg: SupportsCeil):
    assert isinstance(arg, SupportsCeil), f"{arg!r}"


@beartype
def ceil_func_t(arg: SupportsCeilSCU):
    assert isinstance(arg, SupportsCeilSCT), f"{arg!r}"


@beartype
def floor_func(arg: SupportsFloor):
    assert isinstance(arg, SupportsFloor), f"{arg!r}"


@beartype
def floor_func_t(arg: SupportsFloorSCU):
    assert isinstance(arg, SupportsFloorSCT), f"{arg!r}"


# ---- Tests ---------------------------------------------------------------------------


def test_floor_ceil() -> None:
    bool_val: SupportsFloor = True
    int_val: SupportsFloor = -273
    float_val: SupportsFloor = -273.15
    frac_val: SupportsFloor = Fraction(-27315, 100)
    dec_val: SupportsFloor = Decimal("-273.15")
    nw_val: SupportsFloor = Numberwang(-273)
    nwd_val: SupportsFloor = NumberwangDerived(-273)
    nwr_val: SupportsFloor = NumberwangRegistered(-273)
    wn_val: SupportsFloor = Wangernumb(-273.15)
    wnd_val: SupportsFloor = WangernumbDerived(-273.15)
    wnr_val: SupportsFloor = WangernumbRegistered(-273.15)
    _: SupportsCeil
    _ = True
    _ = -273
    _ = -273.15
    _ = Fraction(-27315, 100)
    _ = Decimal("-273.15")
    _ = Numberwang(-273)
    _ = NumberwangDerived(-273)
    _ = NumberwangRegistered(-273)
    _ = Wangernumb(-273.15)
    _ = WangernumbDerived(-273.15)
    _ = WangernumbRegistered(-273.15)

    for good_val in (
        bool_val,
        int_val,
        frac_val,
        dec_val,
        nw_val,
        nwd_val,
        nwr_val,
        wn_val,
        wnd_val,
        wnr_val,
    ):
        assert isinstance(good_val, SupportsFloor), f"{good_val!r}"
        assert isinstance(good_val, SupportsFloorSCT), f"{good_val!r}"
        assert floor(good_val), f"{good_val!r}"
        assert isinstance(good_val, SupportsCeil), f"{good_val!r}"
        assert isinstance(good_val, SupportsCeilSCT), f"{good_val!r}"
        assert ceil(good_val), f"{good_val!r}"

    for out_of_spec_val in (float_val,):
        # Prior to Python 3.9, floats didn't have explicit __floor__ or __ceil__
        # methods; they were "directly" supported in math.floor and math.ceil,
        # respectively, so the pure protocol approach thinks they're not supported
        # TODO(posita): Can we fix this?
        if sys.version_info < (3, 9):
            assert not isinstance(
                out_of_spec_val, SupportsFloor
            ), f"{out_of_spec_val!r}"
            assert not isinstance(out_of_spec_val, SupportsCeil), f"{out_of_spec_val!r}"
        else:
            assert isinstance(out_of_spec_val, SupportsFloor), f"{out_of_spec_val!r}"
            assert floor(out_of_spec_val), f"{out_of_spec_val!r}"
            assert isinstance(out_of_spec_val, SupportsCeil), f"{out_of_spec_val!r}"
            assert ceil(out_of_spec_val), f"{out_of_spec_val!r}"

        # The short-circuiting approach inadvertently (in this case correctly) sweeps in
        # floats, even though they're out-of-spec
        assert isinstance(out_of_spec_val, SupportsFloorSCT), f"{out_of_spec_val!r}"
        assert isinstance(out_of_spec_val, SupportsCeilSCT), f"{out_of_spec_val!r}"

    for bad_val in (
        complex(-273.15),
        "-273.15",
    ):
        assert not isinstance(bad_val, SupportsFloor), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsFloorSCT), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeil), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeilSCT), f"{bad_val!r}"


def test_floor_ceil_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
    ):
        floor_func(cast(SupportsFloor, good_val))
        floor_func_t(cast(SupportsFloorSCU, good_val))
        ceil_func(cast(SupportsCeil, good_val))
        ceil_func_t(cast(SupportsCeilSCU, good_val))

    for out_of_spec_val in (-273.15,):
        # Prior to Python 3.9, floats didn't have explicit __floor__ or __ceil__
        # methods; they were "directly" supported in math.floor and math.ceil,
        # respectively, so the pure protocol approach thinks they're not supported
        # TODO(posita): Can we fix this?
        if sys.version_info < (3, 9):
            with pytest.raises(roar.BeartypeException):
                floor_func(cast(SupportsFloor, out_of_spec_val))

            with pytest.raises(roar.BeartypeException):
                ceil_func(cast(SupportsCeil, out_of_spec_val))
        else:
            floor_func(cast(SupportsFloor, good_val))
            ceil_func(cast(SupportsCeil, good_val))

        # The short-circuiting approach inadvertently (in this case correctly) sweeps in
        # floats, even though they're out-of-spec
        floor_func_t(cast(SupportsFloorSCU, out_of_spec_val))
        ceil_func_t(cast(SupportsCeilSCU, out_of_spec_val))

    for bad_val in (
        complex(-273.15),
        "-273.15",
    ):
        with pytest.raises(roar.BeartypeException):
            floor_func(cast(SupportsFloor, bad_val))

        with pytest.raises(roar.BeartypeException):
            floor_func_t(cast(SupportsFloorSCU, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func(cast(SupportsCeil, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func_t(cast(SupportsCeilSCU, bad_val))


def test_floor_ceil_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: SupportsFloor = numpy.uint8(2)
    uint16_val: SupportsFloor = numpy.uint16(273)
    uint32_val: SupportsFloor = numpy.uint32(273)
    uint64_val: SupportsFloor = numpy.uint64(273)
    int8_val: SupportsFloor = numpy.int8(-2)
    int16_val: SupportsFloor = numpy.int16(-273)
    int32_val: SupportsFloor = numpy.int32(-273)
    int64_val: SupportsFloor = numpy.int64(-273)
    float16_val: SupportsFloor = numpy.float16(-1.8)
    float32_val: SupportsFloor = numpy.float32(-273.15)
    float64_val: SupportsFloor = numpy.float64(-273.15)
    float128_val: SupportsFloor = numpy.float128(-273.15)
    _: SupportsCeil
    _ = numpy.uint8(2)
    _ = numpy.uint16(273)
    _ = numpy.uint32(273)
    _ = numpy.uint64(273)
    _ = numpy.int8(-2)
    _ = numpy.int16(-273)
    _ = numpy.int32(-273)
    _ = numpy.int64(-273)
    _ = numpy.float16(-1.8)
    _ = numpy.float32(-273.15)
    _ = numpy.float64(-273.15)
    _ = numpy.float128(-273.15)

    # numpy.float64 seems to have a closer relationship to the native float than the
    # other numpy.float* types
    for out_of_spec_val in (float64_val,):
        # Prior to Python 3.9, floats didn't have explicit __floor__ or __ceil__
        # methods; they were "directly" supported in math.floor and math.ceil,
        # respectively, so the pure protocol approach thinks they're not supported
        # TODO(posita): Can we fix this?
        if sys.version_info < (3, 9):
            assert not isinstance(
                out_of_spec_val, SupportsFloor
            ), f"{out_of_spec_val!r}"
            assert not isinstance(out_of_spec_val, SupportsCeil), f"{out_of_spec_val!r}"
        else:
            assert isinstance(out_of_spec_val, SupportsFloor), f"{out_of_spec_val!r}"
            assert floor(out_of_spec_val), f"{out_of_spec_val!r}"
            assert isinstance(out_of_spec_val, SupportsCeil), f"{out_of_spec_val!r}"
            assert ceil(out_of_spec_val), f"{out_of_spec_val!r}"

        # The short-circuiting approach inadvertently (in this case correctly) sweeps in
        # floats, even though they're out-of-spec
        assert isinstance(out_of_spec_val, SupportsFloorSCT), f"{out_of_spec_val!r}"
        assert isinstance(out_of_spec_val, SupportsCeilSCT), f"{out_of_spec_val!r}"

    for lying_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
        float16_val,
        float32_val,
        float128_val,
    ):
        # The pure protocol approach catches this
        assert not isinstance(lying_val, SupportsFloor), f"{lying_val!r}"
        assert not isinstance(lying_val, SupportsCeil), f"{lying_val!r}"

        # The short-circuiting approach does not
        assert isinstance(lying_val, SupportsFloorSCT), f"{lying_val!r}"
        assert isinstance(lying_val, SupportsCeilSCT), f"{lying_val!r}"

    for bad_val in (
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert not isinstance(bad_val, SupportsFloor), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsFloorSCT), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeil), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeilSCT), f"{bad_val!r}"


def test_floor_ceil_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    # numpy.float64 seems to have a closer relationship to the native float than the
    # other numpy.float* types
    for out_of_spec_val in (numpy.float64(-273.15),):
        # Prior to Python 3.9, floats didn't have explicit __floor__ or __ceil__
        # methods; they were "directly" supported in math.floor and math.ceil,
        # respectively, so the pure protocol approach thinks they're not supported
        # TODO(posita): Can we fix this?
        if sys.version_info < (3, 9):
            with pytest.raises(roar.BeartypeException):
                floor_func(cast(SupportsFloor, out_of_spec_val))

            with pytest.raises(roar.BeartypeException):
                ceil_func(cast(SupportsCeil, out_of_spec_val))
        else:
            floor_func(cast(SupportsFloor, out_of_spec_val))
            ceil_func(cast(SupportsCeil, out_of_spec_val))

    for lying_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float128(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            floor_func(cast(SupportsFloor, lying_val))

        floor_func_t(cast(SupportsFloorSCU, lying_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func(cast(SupportsCeil, lying_val))

        ceil_func_t(cast(SupportsCeilSCU, lying_val))

    for bad_val in (
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            floor_func(cast(SupportsFloor, bad_val))

        with pytest.raises(roar.BeartypeException):
            floor_func_t(cast(SupportsFloorSCU, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func(cast(SupportsCeil, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func_t(cast(SupportsCeilSCU, bad_val))


def test_floor_ceil_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    integer_val: SupportsFloor = sympy.Integer(-273)
    rational_val: SupportsFloor = sympy.Rational(-27315, 100)
    float_val: SupportsFloor = sympy.Float(-273.15)
    symbol_val: SupportsFloor = sympy.symbols("x")
    _: SupportsCeil
    _ = sympy.Integer(-273)
    _ = sympy.Rational(-27315, 100)
    _ = sympy.Float(-273.15)
    _ = sympy.symbols("x")

    for good_val in (
        integer_val,
        rational_val,
        float_val,
    ):
        assert isinstance(good_val, SupportsFloor), f"{good_val!r}"
        assert isinstance(good_val, SupportsFloorSCT), f"{good_val!r}"
        assert floor(good_val), f"{good_val!r}"
        assert isinstance(good_val, SupportsCeil), f"{good_val!r}"
        assert isinstance(good_val, SupportsCeilSCT), f"{good_val!r}"
        assert ceil(good_val), f"{good_val!r}"

    for bad_val in (symbol_val,):
        assert not isinstance(bad_val, SupportsFloor), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsFloorSCT), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeil), f"{bad_val!r}"
        assert not isinstance(bad_val, SupportsCeilSCT), f"{bad_val!r}"


def test_floor_ceil_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        sympy.Integer(-273),
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
    ):
        floor_func(cast(SupportsFloor, good_val))
        floor_func_t(cast(SupportsFloorSCU, good_val))
        ceil_func(cast(SupportsCeil, good_val))
        ceil_func_t(cast(SupportsCeilSCU, good_val))

    for bad_val in (sympy.symbols("x"),):
        with pytest.raises(roar.BeartypeException):
            floor_func(cast(SupportsFloor, bad_val))

        with pytest.raises(roar.BeartypeException):
            floor_func_t(cast(SupportsFloorSCU, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func(cast(SupportsCeil, bad_val))

        with pytest.raises(roar.BeartypeException):
            ceil_func_t(cast(SupportsCeilSCU, bad_val))



# ======================================================================================
# Copyright and other protections apply. Please see the accompanying LICENSE file for
# rights and restrictions governing use of this software. All rights not expressly
# waived or licensed are reserved. If that file is missing or appears to be modified
# from its original, then please contact the author before viewing or using this
# software in any capacity.
# ======================================================================================

from __future__ import annotations

import math
from decimal import Decimal
from fractions import Fraction
from typing import cast

import pytest

from numerary import IntegralLike, IntegralLikeSCT, IntegralLikeSCU
from numerary.bt import beartype

from .numberwang import (
    Numberwang,
    NumberwangDerived,
    NumberwangRegistered,
    Wangernumb,
    WangernumbDerived,
    WangernumbRegistered,
)

__all__ = ()


# ---- Functions -----------------------------------------------------------------------


@beartype
def func(arg: IntegralLike):
    assert isinstance(arg, IntegralLike)


@beartype
def func_t(arg: IntegralLikeSCU):
    assert isinstance(arg, IntegralLikeSCT)


# ---- Tests ---------------------------------------------------------------------------


def test_integral_like() -> None:
    bool_val: IntegralLike = True
    int_val: IntegralLike = -273
    nw_val: IntegralLike = Numberwang(-273)
    nwd_val: IntegralLike = NumberwangDerived(-273)
    nwr_val: IntegralLike = NumberwangRegistered(-273)

    for good_val in (
        bool_val,
        int_val,
        nw_val,
        nwd_val,
        nwr_val,
    ):
        assert isinstance(good_val, IntegralLike), f"{good_val!r}"
        assert isinstance(good_val, IntegralLikeSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"
        assert good_val - 1 < good_val, f"{good_val!r}"
        assert good_val - 1 <= good_val, f"{good_val!r}"
        assert good_val >= good_val - 1, f"{good_val!r}"
        assert good_val > good_val - 1, f"{good_val!r}"
        assert math.trunc(good_val), f"{good_val!r}"
        assert math.floor(good_val), f"{good_val!r}"
        assert math.ceil(good_val), f"{good_val!r}"
        assert good_val >> 0 == good_val, f"{good_val!r}"
        assert good_val << 0 == good_val, f"{good_val!r}"
        assert good_val & 0 == 0, f"{good_val!r}"
        assert good_val | 0 == good_val, f"{good_val!r}"

    for bad_val in (
        -273.15,
        Fraction(-27315, 100),
        Decimal("-273.15"),
        complex(-273.15),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        assert not isinstance(bad_val, IntegralLike), f"{bad_val!r}"
        assert not isinstance(bad_val, IntegralLikeSCT), f"{bad_val!r}"


def test_integral_like_beartype() -> None:
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        True,
        -273,
        Numberwang(-273),
        NumberwangDerived(-273),
        NumberwangRegistered(-273),
    ):
        func(cast(IntegralLike, good_val))
        func_t(cast(IntegralLikeSCU, good_val))

    for bad_val in (
        -273.15,
        complex(-273.15),
        Fraction(-27315, 100),
        Decimal("-273.15"),
        Wangernumb(-273.15),
        WangernumbDerived(-273.15),
        WangernumbRegistered(-273.15),
        "-273",
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(IntegralLike, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(IntegralLikeSCU, bad_val))


def test_integral_like_numpy() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    uint8_val: IntegralLike = numpy.uint8(2)
    uint16_val: IntegralLike = numpy.uint16(273)
    uint32_val: IntegralLike = numpy.uint32(273)
    uint64_val: IntegralLike = numpy.uint64(273)
    int8_val: IntegralLike = numpy.int8(-2)
    int16_val: IntegralLike = numpy.int16(-273)
    int32_val: IntegralLike = numpy.int32(-273)
    int64_val: IntegralLike = numpy.int64(-273)

    for good_val in (
        uint8_val,
        uint16_val,
        uint32_val,
        uint64_val,
        int8_val,
        int16_val,
        int32_val,
        int64_val,
    ):
        assert isinstance(good_val, IntegralLike), f"{good_val!r}"
        assert isinstance(good_val, IntegralLikeSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"

        # I have no idea why numpy.uint64 is special in this regard
        if isinstance(good_val, numpy.uint64):
            assert good_val >> type(good_val)(0) == good_val, f"{good_val!r}"
            assert good_val << type(good_val)(0) == good_val, f"{good_val!r}"
            assert good_val & type(good_val)(0) == type(good_val)(0), f"{good_val!r}"
            assert good_val | type(good_val)(0) == good_val, f"{good_val!r}"
        else:
            assert good_val >> 0 == good_val, f"{good_val!r}"
            assert good_val << 0 == good_val, f"{good_val!r}"
            assert good_val & 0 == 0, f"{good_val!r}"
            assert good_val | 0 == good_val, f"{good_val!r}"

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        assert not isinstance(bad_val, IntegralLike), f"{bad_val!r}"
        assert not isinstance(bad_val, IntegralLikeSCT), f"{bad_val!r}"


def test_integral_like_numpy_beartype() -> None:
    numpy = pytest.importorskip("numpy", reason="requires numpy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (
        numpy.uint8(2),
        numpy.uint16(273),
        numpy.uint32(273),
        numpy.uint64(273),
        numpy.int8(-2),
        numpy.int16(-273),
        numpy.int32(-273),
        numpy.int64(-273),
    ):
        func(cast(IntegralLike, good_val))
        func_t(cast(IntegralLikeSCU, good_val))

    for bad_val in (
        numpy.float16(-1.8),
        numpy.float32(-273.15),
        numpy.float64(-273.15),
        numpy.float128(-273.15),
        numpy.csingle(-273.15),
        numpy.cdouble(-273.15),
        numpy.clongdouble(-273.15),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(IntegralLike, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(IntegralLikeSCU, bad_val))


def test_integral_like_sympy() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    integral_val: IntegralLike = sympy.Integer(-273)

    for good_val in (integral_val,):
        assert isinstance(good_val, IntegralLike), f"{good_val!r}"
        assert isinstance(good_val, IntegralLikeSCT), f"{good_val!r}"
        assert good_val + 0 == good_val, f"{good_val!r}"
        assert good_val - 0 == good_val, f"{good_val!r}"
        assert good_val * 1 == good_val, f"{good_val!r}"
        assert good_val / 1 == good_val, f"{good_val!r}"
        assert good_val ** 1 == good_val, f"{good_val!r}"
        assert 0 - good_val == -good_val, f"{good_val!r}"
        assert good_val - 1 < good_val, f"{good_val!r}"
        assert good_val - 1 <= good_val, f"{good_val!r}"
        assert good_val >= good_val - 1, f"{good_val!r}"
        assert good_val > good_val - 1, f"{good_val!r}"
        assert math.trunc(good_val), f"{good_val!r}"
        assert math.floor(good_val), f"{good_val!r}"
        assert math.ceil(good_val), f"{good_val!r}"
        assert good_val >> 0 == good_val, f"{good_val!r}"
        assert good_val << 0 == good_val, f"{good_val!r}"
        assert good_val & 0 == 0, f"{good_val!r}"
        assert good_val | 0 == good_val, f"{good_val!r}"

    for bad_val in (
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
        # Bitwise operators are not supported by SymPy's symbols
        sympy.symbols("x"),
    ):
        assert not isinstance(bad_val, IntegralLike), f"{bad_val!r}"
        assert not isinstance(bad_val, IntegralLikeSCT), f"{bad_val!r}"


def test_integral_like_sympy_beartype() -> None:
    sympy = pytest.importorskip("sympy", reason="requires sympy")
    roar = pytest.importorskip("beartype.roar", reason="requires beartype")

    for good_val in (sympy.Integer(-273),):
        func(cast(IntegralLike, good_val))
        func_t(cast(IntegralLikeSCU, good_val))

    for bad_val in (
        sympy.Float(-273.15),
        sympy.Rational(-27315, 100),
        # Bitwise operators are not supported by SymPy's symbols
        sympy.symbols("x"),
    ):
        with pytest.raises(roar.BeartypeException):
            func(cast(IntegralLike, bad_val))

        with pytest.raises(roar.BeartypeException):
            func_t(cast(IntegralLikeSCU, bad_val))
